package rssbuffer

import (
	"log"
	"net/http"
	"strings"
	"sync"
	"time"

	"github.com/SlyMarbo/rss"
)

type Application struct {
	cache *PersistentCache
	log   *log.Logger

	BindTo         string
	DaysOfHistory  int
	RefreshMinutes int
	VerboseLog     bool
}

// NewApplication constructs a new Application object.
func NewApplication(l *log.Logger) (*Application, error) {
	cache, err := NewPersistentCache()
	if err != nil {
		return nil, err
	}

	retn := Application{
		cache: cache,
		log:   l,
	}
	return &retn, nil
}

// feedWorker fetches entries for a given feed.
// As a goroutine it can't return an error, so log them immediately.
func (this *Application) feedWorker(wg *sync.WaitGroup, feed []byte) {
	defer wg.Done()

	if this.VerboseLog {
		this.log.Printf("Refreshing '%s'...\n", string(feed))
	}

	// Network lookup
	f, err := rss.Fetch(string(feed))
	if err != nil {
		this.log.Println(err.Error())
		return
	}

	// Stash feed metadata
	err = this.cache.CacheMetadata(feed, f)
	if err != nil {
		this.log.Println(err.Error())
		return
	}

	// Stash feed items
	for _, itm := range f.Items {
		err = this.cache.CacheEntry(feed, []byte(itm.ID), itm)
		if err != nil {
			this.log.Println(err.Error())
			return
		}
	}
}

// refreshWorker spawns refresh workers for each known RSS feed every 30 minutes
//  (delaying longer if it takes longer to collect all the data).
func (this *Application) refreshWorker() error {
	for {
		wg := sync.WaitGroup{}

		feeds, err := this.cache.GetKnownFeeds()
		if err != nil {
			this.log.Println("Couldn't load feed list: " + err.Error())

		} else {
			for _, feed := range feeds {
				wg.Add(1)
				go this.feedWorker(&wg, feed)
			}
		}

		<-time.After(time.Duration(this.RefreshMinutes) * time.Minute)
		wg.Wait()
	}

	return nil
}

// WebHandler accepts all URLs on our prefix.
func (this *Application) WebHandler(w http.ResponseWriter, r *http.Request) {
	if strings.HasPrefix(r.URL.Path, "/rss/") {
		// suffix is urlencoded
		// FIXME swap to regexp
		var err error = nil
		feed := strings.Replace(strings.Replace(r.URL.Path[5:], "http:/", "http://", -1), "https:/", "https://", -1)

		//feed, err := url.QueryUnescape(r.URL.Path[5:])
		if err != nil {
			w.WriteHeader(400)
			w.Write([]byte("400 BAD REQUEST - couldn't parse URI"))
			return
		}
		feedURIAsByte := []byte(feed)

		feedKnown, err := this.cache.IsKnownFeed(feed)
		if err != nil {
			this.log.Println(err.Error())
			w.WriteHeader(500)
			w.Write([]byte("500 INTERNAL SERVER ERROR"))
			return
		}

		if !feedKnown {
			// New feed, refresh immediately (blocking)
			wg := sync.WaitGroup{}
			wg.Add(1)
			go this.feedWorker(&wg, feedURIAsByte)
			wg.Wait()
		}

		// Determine article cutoff
		cutoff := time.Now().AddDate(0, 0, -this.DaysOfHistory)

		// Look up the metadata and items for the given feed
		feedObject, err := this.cache.Lookup(feedURIAsByte, cutoff)
		if err != nil {
			this.log.Println(err.Error())
			w.WriteHeader(500)
			w.Write([]byte("500 INTERNAL SERVER ERROR"))
			return
		}

		// Produce valid rss output for the given feed
		w.Header().Set("Content-type", "application/rss+xml")
		w.WriteHeader(200)
		BuildRSSDocument(w, feedObject)

		// Done

	} else {
		// 404
		w.WriteHeader(404)
		w.Write([]byte("404 NOT FOUND"))

	}
}

// Run starts the application (blocking)
func (this *Application) Run() {

	this.log.Printf("Listening on %s...", this.BindTo)

	go this.refreshWorker()

	http.HandleFunc("/", this.WebHandler)
	http.ListenAndServe(this.BindTo, http.DefaultServeMux)
}
