package rssbuffer

import (
	"bytes"
	"encoding/gob"
	"errors"
	"time"

	"github.com/SlyMarbo/rss"
	"gopkg.in/boltdb/bolt.v1"
)

const (
	CACHEFILE  = "rssbuffer.cache"
	METABUCKET = "_meta" // mapping of url -> metadata
)

type PersistentCache struct {
	db *bolt.DB
}

func NewPersistentCache() (*PersistentCache, error) {
	db, err := bolt.Open(CACHEFILE, 0600, nil)
	if err != nil {
		return nil, err
	}

	err = db.Update(func(tx *bolt.Tx) error {
		_, err := tx.CreateBucketIfNotExists([]byte(METABUCKET))
		return err
	})
	if err != nil {
		return nil, err
	}

	return &PersistentCache{db: db}, nil
}

// GetKnownFeeds produces a list of feed URLs from the database.
func (this *PersistentCache) GetKnownFeeds() ([][]byte, error) {
	ret := make([][]byte, 0)

	err := this.db.View(func(tx *bolt.Tx) error {
		bucket := tx.Bucket([]byte(METABUCKET))
		c := bucket.Cursor()
		for k, _ := c.First(); k != nil; k, _ = c.Next() {
			k2 := make([]byte, len(k))
			copy(k2, k)
			ret = append(ret, k2)
		}

		return nil
	})

	if err != nil {
		return nil, err
	}

	return ret, nil
}

// IsKnownFeed checks whether a feed is already stored in the database.
func (this *PersistentCache) IsKnownFeed(uri string) (bool, error) {
	knownFeeds, err := this.GetKnownFeeds()
	if err != nil {
		return false, err
	}

	for _, feed := range knownFeeds {
		if string(feed) == uri {
			return true, nil
		}
	}

	return false, nil
}

// CacheMetadata stores the metadata for a given feed.
func (this *PersistentCache) CacheMetadata(feed []byte, data *rss.Feed) error {

	data_minimal := rss.Feed{
		Title:       data.Title,
		Description: data.Description,
		Link:        data.Link,
	}

	return this.db.Update(func(tx *bolt.Tx) error {
		bucket := tx.Bucket([]byte(METABUCKET))

		_, err := tx.CreateBucketIfNotExists(feed) // Also always create feed bucket - ensures we can use a View txn later on
		if err != nil {
			return err
		}

		buff := new(bytes.Buffer)
		enc := gob.NewEncoder(buff)
		err = enc.Encode(data_minimal)
		if err != nil {
			return err
		}

		return bucket.Put(feed, buff.Bytes())
	})
}

// CacheEntry stores the guid/CachedRSSEntry pair for a given feed.
func (this *PersistentCache) CacheEntry(feed, guid []byte, entry *rss.Item) error {
	return this.db.Update(func(tx *bolt.Tx) error {
		bucket := tx.Bucket(feed)

		buff := new(bytes.Buffer)
		enc := gob.NewEncoder(buff)
		err := enc.Encode(entry)
		if err != nil {
			return err
		}

		return bucket.Put(guid, buff.Bytes())
	})
}

// Lookup retrieves the feed metadata and all items up to a time cutoff.
func (this *PersistentCache) Lookup(feed []byte, itemsAfter time.Time) (*rss.Feed, error) {
	retFeed := rss.Feed{}
	retFeed.Items = make([]*rss.Item, 0)

	err := this.db.View(func(tx *bolt.Tx) error {

		// Meta lookup
		meta := tx.Bucket([]byte(METABUCKET))
		feedData := meta.Get(feed)
		if len(feedData) == 0 {
			return errors.New("No data for feed")
		}

		// Unmarshal meta
		err := gob.NewDecoder(bytes.NewReader(feedData)).Decode(&retFeed)
		if err != nil {
			return err
		}

		// Items lookup
		bucket := tx.Bucket(feed)
		c := bucket.Cursor()
		for k, v := c.First(); k != nil; k, v = c.Next() {

			// Unmarshal item
			itm := rss.Item{}
			err = gob.NewDecoder(bytes.NewReader(v)).Decode(&itm)
			if err != nil {
				return err
			}

			// Filter items by date
			if itm.Date.After(itemsAfter) || itm.Date.Equal(itemsAfter) {
				retFeed.Items = append(retFeed.Items, &itm)
			}
		}

		// All done
		return nil
	})
	if err != nil {
		return nil, err
	}

	return &retFeed, nil
}
